/*
 * $Id: protocol.c,v 1.5 2003/03/06 14:05:34 tsavola Exp $
 * Timo Savola <timo.savola@movial.fi>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include "protocol.h"

/**
 * Writes an integer of type int16_t into a file (in network byte order).
 * The integer should be non-negative.
 * @param fd the file descriptor
 * @param i the integer
 * @return 0 on success, -1 on error
 */
int write_int16(int fd,
		int16_t i)
{
	int16_t data = htons(i);

	if (write(fd, &data, sizeof (data)) < 0) {
		return -1;
	}

	return 0;
}

/**
 * Writes a string into a file.
 * @param fd the file descriptor
 * @param str the null-terminated string
 * @return 0 on success, -1 on error
 */
int write_str(int fd,
	      const char *str)
{
	int16_t len;

	len = strlen(str);

	/* Write the length. */
	if (write_int16(fd, len) < 0) {
		return -1;
	}

	/* Write the string. */
	if (write(fd, str, len) < 0) {
		return -1;
	}

	return 0;
}

/**
 * Writes a string vector into a file.
 * @param fd the file descriptor
 * @param strv the string vector
 * @return 0 on success, -1 on error
 */
int write_strv(int fd,
	       const char **strv)
{
	int strc = 0;
	const char **s;

	/* Calculate the length. */
	for (s = strv; *s; s++) {
		strc++;
	}

	/* Write the length. */
	if (write_int16(fd, strc) < 0) {
		return -1;
	}

	/* Write the strings. */
	for (s = strv; *s; s++) {
		if (write_str(fd, *s) < 0) {
			return -1;
		}
	}

	return 0;
}

/**
 * Reads a complete buffer from a file.
 * @param fd a file descriptor
 * @param buf pointer to a buffer large enough
 * @param len the length of the buffer
 * @return 0 on success, -1 on i/o error
 */
int read_buf(int fd,
	     void *buf,
	     int len)
{
	int i, cnt;

	for (i = 0; i < len; i += cnt) {
		cnt = read(fd, buf + i, len - i);
		if (cnt <= 0) {
			return -1;
		}
	}

	return 0;
}

/**
 * Reads an integer of type int16_t from a file (where it is assumed to be
 * in network byte order).
 * @param fd a file descriptor
 * @return -1 on i/o error (thus negative values should not be used)
 */
int16_t read_int16(int fd)
{
	int16_t data;

	if (read_buf(fd, &data, sizeof (data)) < 0) {
		return -1;
	}

	return ntohs(data);
}

/**
 * Reads a string from a file. The returned string should be free()'d.
 * @param fd a file descriptor
 * @return NULL on error
 */
char *read_str(int fd)
{
	int16_t len;
	char *str;

	/* Read the length. */
	len = read_int16(fd);
	if (len < 0) {
		return NULL;
	}

	str = malloc(len + 1);
	if (!str) {
		return NULL;
	}

	/* Read the string. */
	if (read_buf(fd, str, len) < 0) {
		free(str);
		return NULL;
	}
	str[len] = '\0';

	return str;
}

/**
 * Reads a string vector from a file. The returned strings and the vector
 * should be set free(). The vector is null-terminated.
 * @param fd a file descriptor
 * @return NULL on error
 */
char **read_strv(int fd)
{
	int16_t strc;
	char **strv;
	int i;

	/* Read the length. */
	strc = read_int16(fd);
	if (strc < 0) {
		return NULL;
	}

	strv = malloc((strc + 1) * sizeof (char *));
	if (!strv) {
		return NULL;
	}

	/* Read the strings. */
	for (i = 0; i < strc; i++) {
		strv[i] = read_str(fd);
		if (!strv[i]) {
			free(strv);
			return NULL;
		}
	}
	strv[i] = NULL;

	return strv;
}

/**
 * Writes an authentication info structure into a file.
 * @param fd a file descriptor
 * @param user the username
 * @param pwd the password
 * @return 0 on success, -1 on error
 */
int write_auth(int fd,
	       const char *user,
	       const char *pwd)
{
	auth_info_t auth;

	if (strlen(user) >= USER_SIZE) {
		fprintf(stderr, "Warning: username is too long; "
			"truncating to %d characters\n", USER_SIZE - 1);
	}
	if (strlen(pwd) >= PWD_SIZE) {
		fprintf(stderr, "Warning: password is too long; "
			"truncating to %d characters\n", PWD_SIZE - 1);
	}

	memset(auth.user, '\0', USER_SIZE);
	strncpy(auth.user, user, USER_SIZE - 1);

	memset(auth.pwd, '\0', PWD_SIZE);
	strncpy(auth.pwd, pwd, PWD_SIZE - 1);

	if (write(fd, &auth, sizeof (auth)) != sizeof (auth)) {
		return -1;
	}

	return 0;
}

/**
 * Reads an authentication info structure from a file. The returned structure
 * should be free()'d at some point.
 * @param fd a file descriptor
 * @return NULL on error
 */
auth_info_t *read_auth(int fd)
{
	auth_info_t *auth;

	auth = malloc(sizeof (auth_info_t));

	if (read_buf(fd, auth, sizeof (auth_info_t)) < 0) {
		free(auth);
		return NULL;
	}

	return auth;
}

/**
 * Writes a command info structure into a file.
 */
int write_cmd(int fd,
	      const char *target,
	      const char **opts,
	      const char *cmd,
	      const char **argv,
	      const char *curdir,
	      const char **env,
	      int16_t is_tty)
{
	if (write_str  (fd, target) < 0 ||
	    write_strv (fd, opts)   < 0 ||
	    write_str  (fd, cmd)    < 0 ||
	    write_strv (fd, argv)   < 0 ||
	    write_str  (fd, curdir) < 0 ||
	    write_strv (fd, env)    < 0 ||
	    write_int16(fd, is_tty) < 0) {
		return -1;
	}

	return 0;
}

/**
 * Reads a command info structure from a file. The returned structure
 * should be freed with the free_cmd() function.
 * @param fd a file descriptor
 * @return NULL on error
 */
cmd_info_t *read_cmd(int fd)
{
	cmd_info_t *cmd;

	cmd = malloc(sizeof (cmd_info_t));
	memset(cmd, 0, sizeof (cmd_info_t));

	cmd->target = read_str(fd);
	if (!cmd->target) {
		free_cmd(cmd);
		return NULL;
	}

	cmd->opts = read_strv(fd);
	if (!cmd->opts) {
		free_cmd(cmd);
		return NULL;
	}

	cmd->cmd = read_str(fd);
	if (!cmd->cmd) {
		free_cmd(cmd);
		return NULL;
	}

	cmd->argv = read_strv(fd);
	if (!cmd->argv) {
		free_cmd(cmd);
		return NULL;
	}

	cmd->curdir = read_str(fd);
	if (!cmd->curdir) {
		free_cmd(cmd);
		return NULL;
	}

	cmd->env = read_strv(fd);
	if (!cmd->env) {
		free_cmd(cmd);
		return NULL;
	}

	cmd->is_tty = read_int16(fd);
	if (cmd->is_tty < 0) {
		free_cmd(cmd);
		return NULL;
	}

	return cmd;
}

/**
 * Frees a command info structure including its contents.
 */
void free_cmd(cmd_info_t *cmd)
{
	if (cmd) {
		if (cmd->target) {
			free(cmd->target);
		}

		if (cmd->opts) {
			free_vec((void **) cmd->opts);
		}

		if (cmd->cmd) {
			free(cmd->cmd);
		}

		if (cmd->argv) {
			free_vec((void **) cmd->argv);
		}

		if (cmd->curdir) {
			free(cmd->curdir);
		}

		if (cmd->env) {
			free_vec((void **) cmd->env);
		}

		/* Nothing to do about is_tty. */

		free(cmd);
	}
}

/**
 * Writes a packet into a file.
 * @param fd a file descriptor
 * @param type of the packet (one of the PTYPE-values)
 * @param buf the contents of the packet
 * @param size the size of buf
 * @return 0 on success, -1 on i/o error
 */
int write_packet(int fd,
		 uint16_t type,
		 void *buf,
		 uint32_t size)
{
	phead_t head = { type, size };

	if (write(fd, &head, sizeof (head)) < 0) {
		return -1;
	}

	if (size > 0 &&
	    write(fd, buf, size) < 0) {
		return -1;
	}

	return 0;
}

/**
 * Reads the header of a packet from a file so that it can be identified.
 * @param fd a file descriptor
 * @param head buffer for the packet
 * @return 0 on success, -1 on i/o error
 */
int read_phead(int fd,
	       phead_t *head)
{
	return read_buf(fd, head, sizeof (phead_t));
}
